"""
This module defines the Pipe class and related functions for creating and managing pipeline steps.
"""

from buelon.helpers import pipe_util
from . import pipe_interpreter
from . import step_definition
from . import step


class Pipe:
    """
    Represents a pipeline that can create and manage a series of steps.

    Attributes:
        name (str): The name of the pipe.
        imports (str): A comma-separated string of import names.
        kwargs (dict): Optional keyword arguments for the pipe.
    """

    def __init__(self, name: str, imports: str, kwargs: dict = None):
        """
        Initialize a new Pipe instance.

        Args:
            name (str): The name of the pipe.
            imports (str): A comma-separated string of import names.
            kwargs (dict, optional): Keyword arguments for the pipe. Defaults to None.
        """
        self.name = name
        self.imports = imports
        self.kwargs = kwargs or {}

    def create_steps(self, i: int, variables: dict, args: str, env=None) -> list[step.Step]:
        """
        Create a series of steps based on the pipe's imports and given arguments.

        Args:
            i (int): The current line index in the pipeline code.
            variables (dict): A dictionary containing pipeline variables and functions.
            args (str): A comma-separated string of argument names.
            env (optional): Environment variables for the steps. Defaults to None.

        Returns:
            list[step.Step]: A list of created Step instances.

        Raises:
            SyntaxError: If an unknown argument or import is encountered.
            TypeError: If an import is not a StepDefinition instance.
        """
        steps: list[step.Step] = []
        last_step_id = None
        args_step_ids = {}

        if args:
            for arg in args.split(','):
                if arg in variables:
                    variables[arg]: list[step.Step]
                    args_step_ids[arg] = variables[arg][-1].id
                else:
                    raise SyntaxError(f'Line {i+1}: Unknown argument \'{arg}\'')

        for imp in self.imports.split(','):
            if imp not in variables:
                raise SyntaxError(f'Line {i+1}: Unknown import \'{imp}\'')
            if not isinstance(variables[imp], step_definition.StepDefinition):
                raise TypeError(f'Line {i+1}: \'{imp}\' is not an import')

            val: step_definition.StepDefinition = variables[imp]
            _step = val.to_step(env=env)
            if not last_step_id:
                _step.parents = []
                if args:
                    for arg in args.split(','):
                        _step.parents.append(args_step_ids[arg])
                        variables['__steps__'][args_step_ids[arg]].children += [_step.id]
            else:
                _step.parents = [last_step_id]
                variables['__steps__'][last_step_id].children += [_step.id]

            _step.children = []
            last_step_id = _step.id
            variables['__steps__'][_step.id] = _step
            steps.append(_step)

        if steps and not steps[0].parents:
            variables['__starters__'].append(steps[0].id)

        return steps


def check_for_pipe(index: int, line: str, variables: dict) -> bool:
    """
    Check if a line defines a pipe and create a Pipe instance if it does.

    Args:
        index (int): The current line index in the pipeline code.
        line (str): The current line of code being processed.
        variables (dict): A dictionary containing pipeline variables and functions.

    Returns:
        bool: True if a pipe was found and created, False otherwise.

    Raises:
        SyntaxError: If the pipe definition syntax is invalid.
    """
    token = pipe_interpreter.PIPE_TOKENS['pipe']
    if '=' in line.replace('=>', '') and token in line:
        if line.count('=') > 1:
            raise SyntaxError(f'Line {index+1}: Invalid pipe. Too many \'=\'')
        name, imports = line.split('=')
        name = pipe_util.trim(name)
        imports = ','.join([pipe_util.trim(imp) for imp in imports.split(token) if pipe_util.trim(imp) != ''])
        variables[name] = Pipe(name, imports)
        return True
    return False




import os
import asyncio
import traceback
import sys
import enum
import contextlib
from typing import List, Any

import asyncio_pool
import unsync

import buelon.core.step
import buelon.hub
import buelon.bucket
import buelon.helpers.json_parser

try:
    import dotenv
    dotenv.load_dotenv()
except ModuleNotFoundError:
    pass

WORKER_HOST = os.environ.get('PIPE_WORKER_HOST', 'localhost')
WORKER_PORT = int(os.environ.get('PIPE_WORKER_PORT', 65432))
PIPE_WORKER_SUBPROCESS_JOBS = os.environ.get('PIPE_WORKER_SUBPROCESS_JOBS', 'true')

bucket_client = buelon.bucket.Client()
hub_client: buelon.hub.HubClient = buelon.hub.HubClient(WORKER_HOST, WORKER_PORT)

JOB_CMD = f'{sys.executable} -c "import buelon.worker;buelon.worker.job()"'


class HandleStatus(enum.Enum):
    success = 'success'
    pending = 'pending'
    almost = 'almost'
    none = 'none'


@contextlib.contextmanager
def new_client_if_subprocess():
    global hub_client
    if PIPE_WORKER_SUBPROCESS_JOBS == 'true':
        with buelon.hub.HubClient(WORKER_HOST, WORKER_PORT) as client:
            yield client
    else:
        yield hub_client


def job(step_id: str | None = None) -> None:
    if step_id:
        _step = buelon.hub.get_step(step_id)
    else:
        _step = buelon.hub.get_step(os.environ['STEP_ID'])

    if _step is None:
        with new_client_if_subprocess() as client:
            client.reset(step_id if step_id else os.environ['STEP_ID'])
            return

    print('handling', _step.name)
    try:
        args = [buelon.hub.get_data(_id) for _id in _step.parents]
        r: buelon.core.step.Result = _step.run(*args)
        buelon.hub.set_data(_step.id, r.data)

        with new_client_if_subprocess() as client:
            if r.status == buelon.core.step.StepStatus.success:
                client.done(_step.id)
            elif r.status == buelon.core.step.StepStatus.pending:
                client.pending(_step.id)
            elif r.status == buelon.core.step.StepStatus.reset:
                client.reset(_step.id)
            elif r.status == buelon.core.step.StepStatus.cancel:
                client.cancel(_step.id)
            else:
                raise Exception('Invalid step status')
    except Exception as e:
        print(' - Error - ')
        print(str(e))
        traceback.print_exc()
        with new_client_if_subprocess() as client:
            client.error(
                _step.id,
                str(e),
                f'{traceback.format_exc()}'
            )


async def run(step_id: str | None = None) -> None:
    if PIPE_WORKER_SUBPROCESS_JOBS != 'true':
        return job(step_id)
    env = {**os.environ, 'STEP_ID': step_id}
    p = await asyncio.create_subprocess_shell(JOB_CMD, env=env)
    await p.wait()


async def work():
    _scopes: str = os.environ.get('PIPE_WORKER_SCOPES', 'default')
    scopes: List = _scopes.split(',')
    print('scopes', scopes)

    last_loop_had_steps = True

    async with asyncio_pool.AioPool(size=50) as pool:
        while True:
            steps = hub_client.get_steps(scopes)

            if not steps:
                if last_loop_had_steps:
                    last_loop_had_steps = False
                    print('waiting..')
                await asyncio.sleep(1.)
                continue

            last_loop_had_steps = True

            for s in steps:
                await pool.spawn(run(s))


def main():
    asyncio.run(_main())


async def _main():
    global hub_client
    with hub_client:
        await work()


# try:
#     from cython.c_worker import *
# except (ImportError, ModuleNotFoundError):
#     pass


if __name__ == '__main__':
    main()



"""
This module provides utility functions for executing Python code and SQL queries dynamically.

It includes functions for running Python code from strings, executing PostgreSQL and SQLite queries,
and handling dynamic variable substitution in code strings.
"""
from __future__ import annotations
import os
import sys
import importlib
import contextlib
from typing import List, Dict, Any

from buelon.helpers import pipe_util
import buelon.helpers.postgres
import buelon.helpers.sqlite3_helper
from . import pipe_debug


sys.path.append(os.getcwd())


class PipeLineException(Exception):
    """Custom exception for pipeline-related errors."""
    pass


@contextlib.contextmanager
def temp_mod(txt: str):
    """
    Create a temporary Python module from a string.

    Args:
        txt (str): The Python code to be written to the temporary module.

    Yields:
        str: The name of the temporary module.

    Raises:
        PipeLineException: If there's an error in creating or removing the temporary module.
    """
    mod = f'temp_{pipe_util.get_id()}'
    path = os.path.join(os.getcwd(), f'{mod}.py')
    try:
        with open(path, 'w') as f:
            f.write(txt)
        yield mod
    finally:
        try:
            os.remove(path)
        except:
            pass


def fix_data(data: List[Dict]) -> List[Dict]:
    """
    Ensure all dictionaries in a list have the same keys, filling missing keys with None.

    Args:
        data (List[Dict]): A list of dictionaries to fix.

    Returns:
        List[Dict]: The fixed list of dictionaries.
    """
    if isinstance(data, list) and len(data):
        keys = set()
        for row in data:
            keys |= set(row.keys())
        keys = list(data[0].keys()) + list(keys - set(data[0].keys()))
        for row in data:
            for k in keys:
                if k not in row:
                    row[k] = None
    return data


def get_index(k: str) -> tuple[int, str]:
    """
    Extract an index from a string key if present.

    Args:
        k (str): The key string to parse.

    Returns:
        tuple[int, str]: A tuple containing the extracted index (or 0 if not present) and the remaining key.
    """
    if not k.startswith('index['):
        return 0, k
    i, last = '', 0
    for index, c in enumerate(k[len('index['):]):
        last = index
        if c.isdigit():
            i += c
        elif c == ']':
            break
    return int(i), k[last+len('index[')+1:]


def apply_kwargs_to_txt(txt: str, kwargs: dict) -> str:
    """
    Apply keyword arguments to a text string, replacing placeholders.

    Args:
        txt (str): The text containing placeholders.
        kwargs (dict): The keyword arguments to apply.

    Returns:
        str: The text with placeholders replaced by their corresponding values.
    """
    order, skip = {}, {}
    for k, v in kwargs.items():
        i, k = get_index(k)
        if i not in order:
            order[i] = []
        order[i].append((k, v))
        skip[k] = max(skip.get(k, 0), i)
    for i in sorted(order.keys(), reverse=True):
        for k, v in order[i]:
            if skip[k] > i:
                continue
            txt = txt.replace('{{'+k+'}}', f'{v}')
    return txt


def place_null_values(txt: str) -> str:
    """
    Replace unfilled placeholders with a null value string.

    Args:
        txt (str): The text to process.

    Returns:
        str: The processed text with unfilled placeholders replaced.
    """
    return txt  # Currently a no-op, implement as needed


def get_end(txt: str, start: int, end_token: str) -> int:
    """
    Find the end index of a token in a string.

    Args:
        txt (str): The text to search.
        start (int): The starting index for the search.
        end_token (str): The token to search for.

    Returns:
        int: The index of the end token, or -1 if not found.
    """
    for i in range(start, len(txt)):
        if txt[i:i+len(end_token)] == end_token:
            return i
    return -1


def check_for_uuid_kwargs(txt: str, kwargs: dict) -> tuple[str, dict]:
    """
    Check for UUID placeholders in text and replace them with generated UUIDs.

    Args:
        txt (str): The text to process.
        kwargs (dict): Existing keyword arguments.

    Returns:
        tuple[str, dict]: The processed text and a dictionary of new UUID values.

    Raises:
        PipeLineException: If a UUID placeholder is not properly closed.
    """
    n = {}
    for start_token, end_token in [('{{uuid|', '}}'), ('{{uuid:', '}}')]:
        if start_token in txt:
            while start_token in txt:
                s = txt.index(start_token)
                e = get_end(txt, s, end_token)
                if e == -1:
                    raise PipeLineException('uuid not closed')
                _id = txt[s + len(start_token):e]
                key = f'uuid_{_id}'
                if key not in kwargs:
                    kwargs[key] = pipe_util.get_id()
                    n[key] = kwargs[key]
                v = kwargs[key]
                txt = txt[:s] + v + txt[e+2:]
    return txt, n


def check_py_output(data: Any) -> Any:
    """
    Check if the output of a Python execution is of a valid type.

    Args:
        data: The data to check.

    Returns:
        The input data if it's of a valid type.

    Raises:
        PipeLineException: If the data is of an invalid type.
    """
    if not isinstance(data, (list, dict, int, float, str, bool, type(None))):
        raise PipeLineException(f'invalid output type {type(data)} must be: (list, dict, int, float, str, bool, type(None))')
    return data


@pipe_debug.timeit
def run_py(txt: str, func: str, *args, __pure__=False, **kwargs) -> Any:
    """
    Run Python code from a string.

    Args:
        txt (str): The Python code to run.
        func (str): The name of the function to call in the code.
        *args: Positional arguments to pass to the function.
        __pure__ (bool): If True, return the raw output of the function.
        **kwargs: Keyword arguments to apply to the code.

    Returns:
        The result of the function execution, processed based on the __pure__ flag.

    Raises:
        PipeLineException: If the specified function is not found in the code.
    """
    txt = apply_kwargs_to_txt(txt, kwargs)
    txt, uuids = check_for_uuid_kwargs(txt, kwargs)
    txt = place_null_values(txt)

    with temp_mod(txt) as mod:
        module = importlib.import_module(mod)
        if hasattr(module, func):
            r = getattr(module, func)(*args)
            return r
        else:
            raise PipeLineException(f'function {func} not found.')
    return True, []


@pipe_debug.timeit
def run_postgres(txt: str, func: str, *args, **kwargs) -> Any:
    """
    Run a PostgreSQL query.

    Args:
        txt (str): The SQL query to run.
        func (str): The name of the function (used for table naming).
        *args: Data to be uploaded to temporary tables.
        **kwargs: Additional keyword arguments.

    Returns:
        The result of the SQL query execution.

    Raises:
        PipeLineException: If there's an error in query execution.
    """
    db: buelon.helpers.postgres.Postgres = buelon.helpers.postgres.get_postgres_from_env()
    tables = [pipe_util.get_id() for v in args]
    txt, uuids = check_for_uuid_kwargs(txt, kwargs)
    try:
        for i, data in enumerate(args):
            if data:
                n = func if i == 0 else f'pipe{i}'
                if n in txt:
                    try:
                        db.download_table(sql=f'DROP TABLE {tables[i]};')
                    except:
                        pass
                    db.upload_table(tables[i], data)
                    txt = txt.replace(n, tables[i])
        txt = apply_kwargs_to_txt(txt, {**kwargs})
        v = db.download_table(sql=txt)
        return v
    finally:
        for table in tables:
            try:
                db.download_table(f'DROP TABLE {table};')
            except: pass


@pipe_debug.timeit
def run_sqlite3(txt: str, func: str, *args, **kwargs) -> Any:
    """
    Run a SQLite query.

    Args:
        txt (str): The SQL query to run.
        func (str): The name of the function (used for table naming).
        *args: Data to be uploaded to temporary tables.
        **kwargs: Additional keyword arguments.

    Returns:
        The result of the SQL query execution.
    """
    db = buelon.helpers.sqlite3_helper.Sqlite3()
    tables = [pipe_util.get_id() for v in args]
    txt, uuids = check_for_uuid_kwargs(txt, kwargs)
    try:
        for i, data in enumerate(args):
            n = func if i == 0 else f'pipe{i}'
            if n in txt:
                try:
                    db.query(f'DROP TABLE {tables[i]};')
                except:
                    pass
                db.upload_table(tables[i], data)
                txt = txt.replace(n, tables[i])
        txt = apply_kwargs_to_txt(txt, {**kwargs})
        return db.download_table(sql=txt)
    finally:
        try:
            for table in tables:
                db.query(f'DROP TABLE {table};')
        except:
            pass




from . import step
from buelon.helpers import pipe_util
from . import pipe_interpreter


class StepDefinition:
    name: str
    language: str
    function: str
    path: str
    code: str
    using_code: bool = False

    priority: int = 0
    velocity: float = None
    tag: str = None
    scope: str = 'default'

    def __init__(self, name: str, language: str, function: str, path: str, code: str = '', using_code: bool = False, scope: str = None):
        self.name = name
        self.language = language
        self.function = function
        self.path = path
        self.code = code
        self.using_code = using_code
        if scope:
            self.scope = scope

    def to_step(self, env=None) -> step.Step:
        _step = step.Step()
        _step.id = pipe_util.get_id()
        _step.name = self.name
        _step.type = self.language

        if self.using_code:
            _step.code = self.code
        else:
            _step.local = True
            _step.code = self.path
        _step.func = self.function
        _step.kwargs = env or {}
        _step.priority = self.priority
        _step.velocity = self.velocity
        _step.tag = self.tag
        _step.scope = self.scope
        return _step


def check_for_step_definition(index: int, line: str, variables: dict, configurations: dict) -> bool:
    i = index
    if configurations['importing']:
        importing_value: StepDefinition = variables[configurations['current import']]

        if line.strip().startswith('!'):
            if line.count('!') > 1:
                raise SyntaxError(f'Line {i + 1}: Invalid block')
            try:
                importing_value.priority = int(line.strip()[1:])
                return True
            except ValueError:
                raise SyntaxError(f'Line {i + 1}: Invalid priority. Must be integer.')

        if line.strip().startswith('$') and line.strip().count('$') == 1:
            if line.count('$') > 1:
                raise SyntaxError(f'Line {i + 1}: Invalid block')
            importing_value.scope = pipe_util.trim(line.strip()[1:])
            return True

        if line.strip().startswith('@') and line.strip().count('@') == 1:
            if line.count('@') > 1:
                raise SyntaxError(f'Line {i + 1}: Invalid block')
            _line = line.strip()[1:].strip()
            if not _line:
                raise SyntaxError(f'Line {i + 1}: Tag/Velocity cannot be empty')
            if not (set(_line) - set('*/+-. 1234567890')):
                try:
                    importing_value.velocity = eval(_line)
                    return True
                except Exception as e:
                    raise SyntaxError(f'Line {i + 1}: Velocity could not be calculated. ({e})')
            _line = pipe_util.trim(_line)
            if _line not in variables:
                raise SyntaxError(f'Line {i + 1}: Tag/Velocity not found')
            importing_value.tag = pipe_util.trim(_line)
            importing_value.velocity = variables[_line]
            return True

        if not importing_value.language:
            importing_value.language = line.strip()
            if importing_value.language in pipe_interpreter.LANGUAGES:
                importing_value.language = pipe_interpreter.LANGUAGES[importing_value.language]
                return True
            else:
                languages = ', '.join(pipe_interpreter.LANGUAGES.keys())
                raise SyntaxError(
                    f'line {i + 1}: Language {importing_value.language} not found. Languages: {languages}')
        if not importing_value.function:
            importing_value.function = line.strip()
            if not importing_value.function:
                raise SyntaxError(f'line {i + 1}: Function cannot be empty.')
            return True
        if not importing_value.path and not importing_value.using_code:
            def strip_for_block(line):
                return line.strip()  # .replace('\\`', '')

            if not configurations['in code block']:
                if strip_for_block(line).startswith('`'):
                    if strip_for_block(line).endswith('`') and strip_for_block(line).count('`') == 2:
                        importing_value.code = line.strip()[1:-1]
                        importing_value.using_code = True
                        return True
                    elif strip_for_block(line).count('`') > 2:
                        raise SyntaxError(f'Line {i + 1}: Invalid block')
                    importing_value.code = line.strip()[1:]
                    configurations['in code block'] = True
                    return True
                else:
                    importing_value.path = pipe_util.trim(line)
                    return True
            else:
                if strip_for_block(line).endswith('`'):
                    if strip_for_block(line).count('`') > 1:
                        raise SyntaxError(f'Line {i + 1}: Invalid block')
                    importing_value.code += '\n' + line[:line.index('`')]
                    configurations['in code block'] = False
                    importing_value.using_code = True
                    return True
                else:
                    if strip_for_block(line).count('`') > 0:
                        raise SyntaxError(f'Line {i + 1}: Invalid block')
                    importing_value.code += '\n' + line
                    return True
        configurations['importing'] = False

    token = pipe_interpreter.PIPE_TOKENS['import']
    if line.strip().endswith(token) and not line.strip().startswith('for'):
        line = pipe_util.trim(line[:-len(token)])
        variables[line] = StepDefinition(line, '', '', '', '', scope=variables['__scope__'])
        configurations['current import'] = line
        configurations['importing'] = True
        return True
    return False



"""Module for parsing and executing a custom pipeline language.

This module provides functionality to parse and execute code written in a custom
pipeline language. It includes utilities for handling scopes, steps, pipes, actions,
and loops within the custom language.

Constants:
    INDENT (str): A string constant representing four spaces.
    PIPE_TOKENS (dict): A dictionary defining various tokens used in the pipeline language.
    LANGUAGES (dict): A dictionary mapping language names to their standardized identifiers.
"""
import json
import sys
from typing import Dict, Any

from buelon.helpers import pipe_util
from . import step_definition
from . import pipe
from . import action
from . import loop
from . import step


# Constants
INDENT = '    '
PIPE_TOKENS = {
    'import': ':',
    'pipe': '|',
    'func': ('(', ')'),
    'scope': '$',
    'str': ['"'],
    'num': ['#'],
    'bool': ['!'],
    'list': ['[', ']'],
    'call':  ['(', ')'],
    'kwargs': ['{', '}'],
    'block': [{'start': '`', 'end': '`'}]
}


LANGUAGES = {
    'python': step.LanguageTypes.python.value,
    'python3': step.LanguageTypes.python.value,
    'py': step.LanguageTypes.python.value,
    'pg': step.LanguageTypes.postgres.value,
    'postgres': step.LanguageTypes.postgres.value,
    'postgresql': step.LanguageTypes.postgres.value,
    'sqlite3': step.LanguageTypes.sqlite3.value,
    'sqlite': step.LanguageTypes.sqlite3.value
}


def check_for_scope(index: int, line: str, variables: Dict) -> bool:
    """Check if the current line defines a scope and update variables accordingly.

    Args:
        index (int): The current line index.
        line (str): The current line of code.
        variables (dict): A dictionary containing pipeline variables.

    Returns:
        bool: True if a scope was found and processed, False otherwise.

    Raises:
        SyntaxError: If the scope syntax is invalid.
    """
    token = PIPE_TOKENS['scope']
    if line.strip().startswith(token):
        if line.count(token) > 1:
            raise SyntaxError(f'Line {index+1}: Invalid scope')
        scope = line.split(token)[1].strip()
        if not scope:
            raise SyntaxError(f'Line {index+1}: Invalid scope')
        variables['__scope__'] = scope
        return True
    return False


def run(code: str) -> Dict:
    """Parse and execute the given pipeline code.

    Args:
        code (str): The pipeline code to be executed.

    Returns:
        dict: A dictionary containing the parsed pipeline variables.

    Raises:
        Exception: If an error occurs during parsing or execution.
    """
    try:
        variables = {
            '__steps__': {},
            '__starters__': [],
            '__scope__': 'default'
        }

        configurations = {
            'importing': False,
            'current import': None,
            'in loop': False,
            'in code block': False,
        }

        if code.strip().startswith('{'):
            try:
                i, j = pipe_util.extract_json(code)
                kwargs = json.loads(j)
                code = code[i:]
            except Exception as e:
                print(e)

        def remove_comments(line: str) -> str:
            """Remove comments from a line of code.

            Args:
                line (str): A line of code.

            Returns:
                str: The line with comments removed.
            """
            return line.split('#')[0]

        def set_index(_i: int) -> None:
            """Set the current line index.

            This function is used to update the line index, typically when
            processing loops or other constructs that may alter the normal
            flow of line-by-line execution.

            Args:
                _i (int): The new line index to set.
            """
            nonlocal i
            i = _i

        def read_line(index: int) -> None:
            """Process a single line of the pipeline code.

            Args:
                index (int): The index of the current line.
            """
            nonlocal i
            line = lines[index]

            # else:
            if step_definition.check_for_step_definition(i, line, variables, configurations):
                return

            if check_for_scope(i, line, variables):
                return

            if pipe.check_for_pipe(i, line, variables):
                return

            if action.check_for_actions(i, line, variables):
                return

            if loop.check_loop(i, line, lines, variables, read_line, set_index):
                return

            if line.strip():
                raise SyntaxError(f'Line {i+1}: Invalid syntax')

        lines = [remove_comments(line) for line in code.splitlines()]
        i = 0
        while i < len(lines):
            read_line(i)
            i += 1

        return variables

    except Exception as e:
        print(f'Error {type(e).__name__} | {e}')
        sys.exit(1)


def get_steps_from_code(code: str) -> Dict[str, Any]:
    """Extract steps and starters from the given pipeline code.

    Args:
        code (str): The pipeline code to be processed.

    Returns:
        dict: A dictionary containing the extracted steps and starters.
    """
    variables = run(code)
    return {'steps': variables['__steps__'], 'starters': variables['__starters__']}




from setuptools import setup, find_packages, Extension
from Cython.Build import cythonize

# List of Cython files to compile
extensions = [
    Extension("buelon.cython.c_bucket", ["buelon/cython/c_bucket.pyx"]),
    Extension("buelon.cython.c_worker", ["buelon/cython/c_worker.pyx"]),
    Extension("buelon.cython.c_hub", ["buelon/cython/c_hub.pyx"])
]

# Requirements  for the package
requirements = [
    'Cython',
    'psycopg2-binary',
    'orjson',
    'python-dotenv',
    'asyncio-pool',
    'psutil',
    'unsync',
    'redis'
]

# Read the long description from the README file
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="buelon",
    version="1.0.15",
    author="Daniel Olson",
    author_email="daniel@orphos.cloud",
    description="A scripting language to simply manage a very large amount of i/o heavy workloads. Such as API calls "
                "for your ETL, ELT or any program needing Python and/or SQL",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/daniel-olson-code/buelon",
    packages=find_packages(),
    package_data={
        'buelon/examples': [
            "example.pipe",
        ],
        'buelon/cython': [
            "*.pyx",
        ]
    },
    include_package_data=True,
    package_name="buelon",
    ext_modules=cythonize(extensions),
    install_requires=requirements,
    entry_points={
        'console_scripts': ['bue=buelon.command_line:cli'],
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    python_requires=">=3.10",
    keywords="buelon etl pipeline asynchronous data-processing api",
)


import asyncio
import enum
import traceback
import sqlite3
import os
import socket
import threading
import queue
import time
from typing import Any

import unsync
try:
    import dotenv
    dotenv.load_dotenv()
except ModuleNotFoundError:
    pass

import buelon.bucket
import buelon.core.step
import buelon.core.pipe_interpreter
import buelon.helpers.json_parser
import buelon.core.pipe_debug


class Method(enum.Enum):
    GET_STEPS = 'get-steps'
    DONE = 'done'
    PENDING = 'pending'
    CANCEL = 'cancel'
    RESET = 'reset'
    ERROR = 'error'
    UPLOAD_STEP = 'upload-step'
    STEP_COUNT = 'step-count'
    RESET_ERRORS = 'reset-errors'
    DELETE_STEPS = 'delete-steps'


PIPELINE_HOST = os.environ.get('PIPELINE_HOST', '0.0.0.0')
PIPELINE_PORT = int(os.environ.get('PIPELINE_PORT', 65432))

db_path = os.path.join('database.db')

bucket_client = buelon.bucket.Client()

# Initialize a global tag_usage dictionary
tag_usage = {}
tag_lock = threading.Lock()

# Create a queue to handle incoming connections
connection_queue = queue.Queue()

PIPELINE_SPLIT_TOKEN = b'|-**-|'
PIPELINE_END_TOKEN = b'[-_-]'
LENGTH_OF_PIPELINE_END_TOKEN = len(PIPELINE_END_TOKEN)


def receive(conn: socket.socket) -> bytes:
    data = b''
    while not data.endswith(PIPELINE_END_TOKEN):
        v = conn.recv(1024)
        data += v
    return data[:-LENGTH_OF_PIPELINE_END_TOKEN]


def send(conn: socket.socket, data: bytes) -> None:
    conn.sendall(data+PIPELINE_END_TOKEN)


def load_db():
    with sqlite3.connect(db_path) as conn:
        cur = conn.cursor()
        cur.execute('PRAGMA journal_mode=WAL')
        cur.execute('CREATE TABLE IF NOT EXISTS steps ('
                    'id, priority, scope, velocity, tag, status, epoch, msg, trace);')
        cur.execute('CREATE TABLE IF NOT EXISTS tags ('
                    'tag, velocity);')
        conn.commit()


def delete_steps():
    WORKER_HOST = os.environ.get('PIPE_WORKER_HOST', 'localhost')
    WORKER_PORT = int(os.environ.get('PIPE_WORKER_PORT', 65432))
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        s.connect((WORKER_HOST, WORKER_PORT))
        data = (b'delete-steps'
                + buelon.hub.PIPELINE_SPLIT_TOKEN
                + b'nothing')
        send(s, data)


def _delete_steps():
    with sqlite3.connect(db_path) as conn:
        cur = conn.cursor()
        cur.execute('delete from steps;')
        conn.commit()


def reset_errors(include_workers=False):
    WORKER_HOST = os.environ.get('PIPE_WORKER_HOST', 'localhost')
    WORKER_PORT = int(os.environ.get('PIPE_WORKER_PORT', 65432))
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        s.connect((WORKER_HOST, WORKER_PORT))
        data = (b'reset-errors'
                + buelon.hub.PIPELINE_SPLIT_TOKEN
                + (b'true' if include_workers else b'false'))
        send(s, data)


def _reset_errors(include_workers=b'false'):
    suffix = '' if include_workers != b'true' else f' or status = \'{buelon.core.step.StepStatus.working.value}\''
    query = f'''
    update steps
    set status = \'{buelon.core.step.StepStatus.pending.value}\'
    where status = \'{buelon.core.step.StepStatus.error.value}\'
    {suffix};'''
    with sqlite3.connect(db_path) as conn:
        cur = conn.cursor()
        cur.execute(query)
        conn.commit()


def get_step_count(types: str | None = None) -> list[dict]:
    WORKER_HOST = os.environ.get('PIPE_WORKER_HOST', 'localhost')
    WORKER_PORT = int(os.environ.get('PIPE_WORKER_PORT', 65432))
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        s.connect((WORKER_HOST, WORKER_PORT))
        data = (b'step-count'
                + buelon.hub.PIPELINE_SPLIT_TOKEN
                + buelon.helpers.json_parser.dumps({'types': types}))
        send(s, data)
        return buelon.helpers.json_parser.loads(receive(s))


def _get_step_count(types: str | None = None) -> list[dict]:
    if types == '*':
        where = ''
    else:
        where = f'''
        where
            status not in (
                '{buelon.core.step.StepStatus.success.value}',
                '{buelon.core.step.StepStatus.cancel.value}'
            )
        '''

    query = f'''
    select
        status,
        count(*) as amount
    from
        steps
    {where}
    group by
        status;
    '''
    with sqlite3.connect(db_path) as conn:
        cur = conn.cursor()
        cur.execute(query)

        headers = [row[0] for row in cur.description]
        table = [dict(zip(headers, row)) for row in cur.fetchall()]

        for row in table:
            row['status'] = buelon.core.step.StepStatus(int(row['status'])).name

    return table


def upload_step(_step: buelon.core.step.Step, status: buelon.core.step.StepStatus) -> None:
    WORKER_HOST = os.environ.get('PIPE_WORKER_HOST', 'localhost')
    WORKER_PORT = int(os.environ.get('PIPE_WORKER_PORT', 65432))
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        s.connect((WORKER_HOST, WORKER_PORT))
        data = (b'upload-step'
                + buelon.hub.PIPELINE_SPLIT_TOKEN
                + buelon.helpers.json_parser.dumps([_step.to_json(), status.value]))
        send(s, data)


def _upload_step(step_json: dict, status_value: int) -> None:
    status = buelon.core.step.StepStatus(status_value)
    _step = buelon.core.step.Step().from_json(step_json)

    with sqlite3.connect(db_path) as conn:
        cur = conn.cursor()
        sql = ('INSERT INTO steps (id, priority, scope, velocity, tag, status, epoch, msg, trace) '
               'VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?);')
        cur.execute(sql, (_step.id, _step.priority, _step.scope, _step.velocity, _step.tag, f'{status.value}',
                          time.time(), '', ''))
        conn.commit()


def upload_pipe_code(code: str):
    client = HubClient()

    variables = buelon.core.pipe_interpreter.get_steps_from_code(code)
    steps = variables['steps']
    starters = variables['starters']
    print('uploading', len(steps), 'steps')
    add_later = []
    for _step in steps.values():
        if _step.id in starters:
            add_later.append(_step)
            continue
        set_step(_step)
        # status = buelon.core.step.StepStatus.pending if _step.id in starters else buelon.core.step.StepStatus.queued
        # upload_step(_step, status)
        # upload_step(_step, buelon.core.step.StepStatus.queued)
        client.sync_upload_step(_step.to_json(), buelon.core.step.StepStatus.queued.value)

    for _step in add_later:
        set_step(_step)
        # upload_step(_step, buelon.core.step.StepStatus.pending)
        client.sync_upload_step(_step.to_json(), buelon.core.step.StepStatus.pending.value)


def upload_pipe_code_from_file(file_path: str):
    with open(file_path, 'r') as f:
        code = f.read()
        upload_pipe_code(code)


def get_step(step_id: str) -> buelon.core.step.Step:
    s = buelon.core.step.Step()
    b = bucket_client.get(f'step/{step_id}')
    data = buelon.helpers.json_parser.loads(b)
    return s.from_json(data)


def set_step(_step: buelon.core.step.Step) -> None:
    b = buelon.helpers.json_parser.dumps(_step.to_json())
    bucket_client.set(f'step/{_step.id}', b)


def get_data(step_id: str) -> Any:
    key = f'step-data/{step_id}'
    v = bucket_client.get(key)
    if v is None:
        raise ValueError(f'No data found for step {step_id}')
    return buelon.helpers.json_parser.loads(v)


def set_data(step_id: str, data: Any) -> None:
    key = f'step-data/{step_id}'
    b = buelon.helpers.json_parser.dumps(data)
    bucket_client.set(key, b)


def remove_data(step_id: str) -> None:
    key = f'step-data/{step_id}'
    bucket_client.delete(key)


def check_to_delete_bucket_files(
        step_id: str,
        already: set | None = None,
        steps: list | None = None
) -> None:
    _check_to_delete_bucket_files(step_id, already, steps)


def _check_to_delete_bucket_files(
        step_id: str,
        already: set | None = None,
        steps: list | None = None
) -> None:
    first_iteration = already is None and steps is None
    _step = get_step(step_id)
    already = set() if first_iteration else already
    steps = [] if first_iteration else steps
    already.add(_step.id)
    steps.append(_step)

    if _step.parents:
        for parent in _step.parents:
            if parent not in already:
                already.add(parent)
                check_to_delete_bucket_files(parent, already, steps)

    if _step.children:
        for child in _step.children:
            if child not in already:
                already.add(child)
                check_to_delete_bucket_files(child, already, steps)

    if first_iteration:
        ids = [s.id for s in steps]
        finished_statuses = {f'{v}' for v in [buelon.core.step.StepStatus.cancel.value, buelon.core.step.StepStatus.success.value]}
        with sqlite3.connect(db_path) as conn:
            cur = conn.cursor()
            sql = f'SELECT status FROM steps WHERE id IN ({", ".join("?" * len(ids))})'
            cur.execute(sql, ids)
            rows = cur.fetchall()

            if all([f'{row[0]}' in finished_statuses for row in rows]):
                for s in steps:
                    remove_data(s.id)


def _done(step_id: str):
    with sqlite3.connect(db_path) as conn:
        cur = conn.cursor()
        _step = get_step(step_id)
        # set_data(_step.id, result.data)
        sql_update_step = (f'UPDATE steps SET status = \'{buelon.core.step.StepStatus.success.value}\', epoch = ? WHERE id = ?')
        cur.execute(sql_update_step, (time.time(), _step.id))

        # cur.execute(sql_set_status, (_step.id, ))
        conn.commit()

        if _step.children:
            sql_update_children = (f'UPDATE steps SET status = \'{buelon.core.step.StepStatus.pending.value}\', epoch = ? WHERE id IN ({", ".join("?" * len(_step.children))})')
            cur.execute(sql_update_children, (time.time(), *_step.children))
            conn.commit()

        check_to_delete_bucket_files(step_id)


def _pending(step_id: str):
    with sqlite3.connect(db_path) as conn:
        cur = conn.cursor()
        sql_update_step = (f'UPDATE steps SET status = \'{buelon.core.step.StepStatus.pending.value}\', epoch = ? WHERE id = ?')
        cur.execute(sql_update_step, (time.time(), step_id))


def _cancel(
        step_id: str,
        already: set | None = None
) -> None:
    first_iteration = already is None
    _step = get_step(step_id)
    already = set() if first_iteration else already

    sql_update_step = f'UPDATE steps SET status = \'{buelon.core.step.StepStatus.cancel.value}\', epoch = ? WHERE id = ?'

    with sqlite3.connect(db_path) as conn:
        cur = conn.cursor()
        cur.execute(sql_update_step, (time.time(), _step.id))
        conn.commit()

    if _step.parents:
        for parent in _step.parents:
            if parent not in already:
                already.add(parent)
                _cancel(parent, already)

    if _step.children:
        for child in _step.children:
            if child not in already:
                already.add(child)
                _cancel(child, already)

    if first_iteration:
        check_to_delete_bucket_files(step_id)


def _reset(step_id: str, already=None):
    _step = get_step(step_id)
    already = set() if not already else already
    status = buelon.core.step.StepStatus.pending.value if _step.parents else buelon.core.step.StepStatus.queued.value
    sql_update_step = f'UPDATE steps SET status = \'{status}\', epoch = ? WHERE id = ?'

    with sqlite3.connect(db_path) as conn:
        cur = conn.cursor()
        cur.execute(sql_update_step, (time.time(), _step.id))
        conn.commit()

    if _step.children:
        for child in _step.children:
            if child not in already:
                already.add(child)
                _reset(child, already)

    if _step.parents:
        for parent in _step.parents:
            if parent not in already:
                already.add(parent)
                _reset(parent, already)


def _error(step_id: str, msg: str, trace: str):
    _step = get_step(step_id)
    sql_update_step = (f'UPDATE steps SET status = \'{buelon.core.step.StepStatus.error.value}\', epoch = ?, msg = ?, trace = ? WHERE id = ?')

    with sqlite3.connect(db_path) as conn:
        cur = conn.cursor()
        cur.execute(sql_update_step, (time.time(), msg, trace, _step.id))
        conn.commit()


def get_steps(scopes: list, limit=50, chunk_size=100):
    """
    Get the steps in the scope, ordered by priority, velocity, then scope position.

    Args:
        scopes (list): A list of scopes to filter the steps.
        limit (int): The maximum number of steps to retrieve. Defaults to 50.
        chunk_size (int): The number of steps to fetch in each chunk. Defaults to 100.

    Returns:
        list: A list of steps ordered by priority, velocity, and scope position.
    """
    global tag_usage

    with sqlite3.connect(db_path) as conn:
        cur = conn.cursor()
        # Fetch velocities for each tag
        velocity_sql = 'SELECT tag, velocity FROM tags'  # f'SELECT tag, velocity FROM tags WHERE scope IN ({",".join("?" * len(scopes))})'
        cur.execute(velocity_sql)  # (velocity_sql, (*scopes,))
        tag_velocities = dict(cur.fetchall())

        # Initialize tag_usage for new tags
        for tag in tag_velocities:
            if tag not in tag_usage:
                tag_usage[tag] = 0

        case_statement = ' '.join([f"WHEN ? THEN {i}" for i in range(len(scopes))])
        offset = 0
        steps = []

        while len(steps) < limit:
            sql = (f'SELECT id, priority, scope, velocity, tag '
                   f'FROM steps '
                   f'WHERE scope IN ({",".join("?" * len(scopes))}) '
                   f'AND ('
                   f'   status = \'{buelon.core.step.StepStatus.pending.value}\' '
                   f'   or (epoch < {time.time() - (60 * 60 * 2)} and status = \'{buelon.core.step.StepStatus.working.value}\')'
                   f')'
                   f'ORDER BY CASE scope {case_statement} END, priority desc, epoch '  # , COALESCE(velocity, 1.0/0.0)
                   f'LIMIT ? OFFSET ?')
            cur.execute(sql, (*scopes, *scopes, chunk_size, offset))
            rows = cur.fetchall()
            if not rows:
                break  # Exit loop if no more rows are fetched

            for row in rows:
                step_id, priority, scope, velocity, tag = row
                if tag not in tag_velocities:
                    tag_velocities[tag] = None
                    tag_usage[tag] = 0
                if tag_velocities[tag] is None or tag_usage[tag] < tag_velocities[tag]:
                    steps.append(step_id)  # (get_step(step_id))
                    tag_usage[tag] += 1
                    if len(steps) >= limit:
                        break

            offset += chunk_size

        sql_set_to_working = f'UPDATE steps SET status = \'{buelon.core.step.StepStatus.working.value}\', epoch = {time.time()} WHERE id IN ({", ".join("?" * len(steps))})'
        cur.execute(sql_set_to_working, steps)
        conn.commit()

    return steps


def reset_tag_usage():
    global tag_usage
    tag_usage = {}


# Function to decrement tag usage every second

def decrement_tag_usage():
    global tag_usage
    while True:
        time.sleep(1)  # Sleep for 1 second

        with tag_lock:
            for tag in list(tag_usage.keys()):  # Use list() to create a copy of keys for safe iteration
                tag_usage[tag] = max(0, tag_usage[tag] - 1)

class HubServer:
    def __init__(self, host='0.0.0.0', port=65432):
        self.host = host
        self.port = port
        self.transaction_queue = queue.Queue()

    def start(self):
        server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server.bind((self.host, self.port))
        server.listen()
        print(f"Server listening on {self.host}:{self.port}")

        worker_thread = threading.Thread(target=self._process_transactions)
        worker_thread.start()

        while True:
            client_socket, addr = server.accept()
            client_thread = threading.Thread(target=self._handle_client, args=(client_socket,))
            client_thread.start()

    def _handle_client(self, client_socket):
        with client_socket:
            try:
                while True:
                    data = receive(client_socket)

                    if not data:
                        break

                    method, payload = data.split(PIPELINE_SPLIT_TOKEN)
                    print('received', method)
                    method = Method(method.decode())
                    self.transaction_queue.put((method, payload, client_socket))
            except KeyboardInterrupt:
                raise
            except Exception as e:
                print(f"Error handling client: {e}")
                traceback.print_exc()

    def _process_transactions(self):
        while True:
            method, payload, client_socket = self.transaction_queue.get()
            try:
                print('processing', method)
                response = self._process_request(method, payload)

                send(client_socket, response)

            except Exception as e:
                traceback.print_exc()
                print(f"Error processing transaction: {e}")
            finally:
                self.transaction_queue.task_done()

    def _process_request(self, method: Method, payload: bytes):
        if method == Method.GET_STEPS:
            scopes = buelon.helpers.json_parser.loads(payload)
            steps = get_steps(scopes)
            return buelon.helpers.json_parser.dumps(steps)
        elif method == Method.DONE:
            _done(payload.decode())
        elif method == Method.PENDING:
            _pending(payload.decode())
        elif method == Method.CANCEL:
            _cancel(payload.decode())
        elif method == Method.RESET:
            _reset(payload.decode())
        elif method == Method.ERROR:
            values = buelon.helpers.json_parser.loads(payload)
            _error(values['step_id'], values['msg'], values['trace'])
        elif method == Method.UPLOAD_STEP:
            step_json, status_value = buelon.helpers.json_parser.loads(payload)
            _upload_step(step_json, status_value)
        elif method == Method.STEP_COUNT:
            kwargs = buelon.helpers.json_parser.loads(payload)
            result = _get_step_count(kwargs['types'])
            return buelon.helpers.json_parser.dumps(result)
        elif method == Method.RESET_ERRORS:
            _reset_errors(payload)
        elif method == Method.DELETE_STEPS:
            _delete_steps()
        return b'ok'


class HubClient:
    def __init__(self, host=None, port=None, max_reconnect_attempts=2):
        self.host = host or os.environ.get('PIPE_WORKER_HOST', 'localhost')
        self.port = port or int(os.environ.get('PIPE_WORKER_PORT', 65432))
        self.conn = None
        self.max_reconnect_attempts = max_reconnect_attempts

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def connect(self):
        if not self.conn:
            self.conn = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.conn.connect((self.host, self.port))

    def close(self):
        if self.conn:
            self.conn.close()
            self.conn = None

    def send_request(self, method, data: bytes):
        attempt = 0
        while attempt < self.max_reconnect_attempts:
            try:
                if not self.conn:
                    self.connect()
                request = method.value.encode() + PIPELINE_SPLIT_TOKEN + data
                send(self.conn, request)
                response = receive(self.conn)
                return response
            except (OSError, ConnectionError) as e:
                print(f"Connection error: {e}")
                attempt += 1
                self.close()
                if attempt >= self.max_reconnect_attempts:
                    raise ConnectionError(f"Failed to send request after {self.max_reconnect_attempts} attempts: {e}")
                time.sleep(1)  # Backoff before retrying
                self.connect()  # Establish a new connection before retrying

    def get_steps(self, scopes):
        return buelon.helpers.json_parser.loads(self.send_request(Method.GET_STEPS, buelon.helpers.json_parser.dumps(scopes)))

    def done(self, step_id):
        return self.send_request(Method.DONE, step_id.encode())

    def pending(self, step_id):
        return self.send_request(Method.PENDING, step_id.encode())

    def cancel(self, step_id):
        return self.send_request(Method.CANCEL, step_id.encode())

    def reset(self, step_id):
        return self.send_request(Method.RESET, step_id.encode())

    def error(self, step_id, msg, trace):
        return self.send_request(Method.ERROR, buelon.helpers.json_parser.dumps({'step_id': step_id, 'msg': msg, 'trace': trace}))

    def upload_step(self, step_json, status_value):
        return self.send_request(Method.UPLOAD_STEP, buelon.helpers.json_parser.dumps([step_json, status_value]))

    def get_step_count(self, types: str | None = None):
        return buelon.helpers.json_parser.loads(
            self.send_request(
                Method.STEP_COUNT,
                buelon.helpers.json_parser.dumps({'types': types})
            )
        )

    def reset_errors(self, include_workers):
        return self.send_request(Method.RESET_ERRORS, b'true' if include_workers else b'false')

    def delete_steps(self):
        return self.send_request(Method.DELETE_STEPS, b'nothing')

    def __getattr__(self, item: str):
        if item.startswith('sync_'):
            if hasattr(self, item[5:]):
                return getattr(self, item[5:])
        raise AttributeError('')


def main():
    """
    Main function to start the server and worker thread.

    Handles keyboard interruption to shut down the server gracefully.
    """
    load_db()

    server = HubServer()
    server.start()


if __name__ == "__main__":
    main()
